package BoGThomsenPlayer;

import java.util.ArrayList;

import Player.Player;

/**
 * Automatic computer player for Connect4 game. 
 * The player calculates 5 moves ahead using minimax search 
 * 
 * @author Bo G
 */
public class BoGThomsenPlayer implements Player{

	Boolean color;
	AIPlayerBoard b;
	private static final int NUM_OF_MOVES_TO_LOOK_AHEAD = 5;

	/**
	 * Init player by setting player color
	 * @param color The player color. true is red, false is yellow
	 */
	public void init(Boolean color){
		b = new AIPlayerBoard();
		this.color = color;
	}

	/**
	 * Return the player name i.e. the class name
	 */
	public String name(){
		return "BoGThomsenPlayer";
	}

	/**
	 * Get the next move
	 */
	public int move(){
		if (!b.isGameOver()) {
			// First 2 moves hardcoded
			if (b.getMovesRemaining() == 42){
				b.makeMove(2, color);
				return 2;
			}
			else if (b.getMovesRemaining() == 41){
				if (b.isColumnEmpty(3)){
					b.makeMove(3, color);
					return 3;
				} else { 
					b.makeMove(2, color);
					return 2;
				}
			}

			MiniMaxCalculator calc = new MiniMaxCalculator();
			int moveToMake = calc.getNextMove(NUM_OF_MOVES_TO_LOOK_AHEAD);

			b.makeMove(moveToMake, color);
			return moveToMake;
		} else 
			return -1;

	}

	/**
	 * Informs the player that the opponent has made the indicated move
	 */
	public void inform(int i){
		if (color)
			b.makeMove(i, false);
		else
			b.makeMove(i, true);
	}

///////////// Field  ///////////

	/**
	 * A class that represents 1 of the 42 fields on the Connect4 game board
	 * Holds only the color of the player owning the field
	 * @author Bo G
	 */
	public class AIPlayerField {
		
		private Boolean color = null;
		
		public AIPlayerField(){}

		/**
		 * Get the color of this Field
		 * @return True if Red, False if yellow and null if empty
		 */
		public Boolean getColor(){
			return color;
		}
		/**
		 * Set the color of this Field
		 * @param color True is red, False is yellow
		 */
		public void setColor(Boolean color){
			this.color = color;
		}
	}


/////////// Combination ///////////

	/**
	 * A Class that represents a 4 in a row combination
	 * @author Bo G
	 */
	public class AIPlayerCombination {

		private AIPlayerField f1;
		private AIPlayerField f2;
		private AIPlayerField f3;
		private AIPlayerField f4;

		/**
		 * Constructor
		 * @param f1 The first Field in the combination
		 * @param f2 The second Field in the combination
		 * @param f3 The third Field in the combination
		 * @param f4 The fourth Field in the combination
		 */
		public AIPlayerCombination(AIPlayerField f1, AIPlayerField f2, AIPlayerField f3, AIPlayerField f4){
			this.f1 = f1;
			this.f2 = f2;
			this.f3 = f3;
			this.f4 = f4;
		}

		/** 
		 * Test if this is a wining combination
		 * @param color The color to test with
		 * @return True if all Fields are the same color as color
		 */
		public boolean isWinningCombo(Boolean color){
			if (f1.getColor() != null){
				if (f1.getColor() == color &&
						f2.getColor() == color &&
						f3.getColor() == color &&
						f4.getColor() == color)
					return true;
				else 
					return false;
			} else
				return false;
		}

		/**
		 * Check whether combo is a nth in a row. This is true if the combination has 'howMany' player owned 
		 * fields in it and no fields belonging to the second player 
		 * @param color The player color
		 * @param howMany The number in a row (2, 3 or 4)
		 * @return True if combination has nth number of player fields in it and no hostile fields
		 */
		public boolean isNthInARow(Boolean color, int howMany){

			int count = 0;
			int aim = howMany;
			Boolean adversary = true;
			if (color)
				adversary = false;

			if (f1.getColor() == adversary ||
					f2.getColor() == adversary ||
					f3.getColor() == adversary ||
					f4.getColor() == adversary)
				return false;
			else {
				if (f1.getColor() == color) count++;
				if (f2.getColor() == color) count++;
				if (f3.getColor() == color) count++;
				if (f4.getColor() == color) count++;

				if (count == aim)
					return true;
				else 
					return false;
			}
		}

		/**
		 * Check if this combination contains the specified field
		 * @param f The field to test with
		 * @return True if Field is in this combination
		 */
		public boolean containsField(AIPlayerField f){
			if (f1 == f || f2 ==f || f3 == f ||f4 == f)
				return true;
			else 
				return false;				
		}
	}	

/////////// Game state ///////////

	/**
	 * A class that represents the state of the current game
	 * @author Bo G
	 */
	public class AIPlayerGameState {

		private boolean gameOver;
		private int movesRemaining;
		private Boolean winningPlayer;
		private int playerToMove;

		/**
		 * Constructor initializes the values
		 */
		public AIPlayerGameState(){
			gameOver = false;
			movesRemaining = 42;
			playerToMove = 0;
			winningPlayer = null;
		}

		/**
		 * Is this game over?
		 * @return True if game over
		 */
		public boolean isGameOver(){
			return gameOver;
		}

		/**
		 * Returns the moves remaining 
		 * @return Number of moves remaining
		 */
		public int getMovesRemaining(){
			return movesRemaining;
		}

		/**
		 * Decrease the amount of moves remaining and set the next player to move
		 */
		public void decreaseMovesRemaining(){		
			if (--movesRemaining == 0)
				gameOver = true;
			else
				setNextPlayerToMove();
		}

		/**
		 * Set next player to move.
		 */
		public void setNextPlayerToMove(){
			if (playerToMove == 0)
				playerToMove = 1;
			else
				playerToMove = 0;
		}

		/**
		 * Set the color of the player who won
		 * @param color The color of the winning player
		 */
		public void setPlayerWon(Boolean color){
			winningPlayer = color;
		}

		/**
		 * Set the gameOver property
		 * @param gameOver True if game is finished
		 */
		public void setGameOver(boolean gameOver){
			this.gameOver = gameOver;
		}

		/** 
		 * Increase the number of moves remaining by 1. Used in conjunction with undo.
		 */
		public void increaseMovesRemaining(){
			movesRemaining++;
		}
	}

/////////// Game board ///////////

	/**
	 * A class that represents a Connect4 game board
	 * @author Bo G
	 */
	public class AIPlayerBoard {

		private ArrayList<ArrayList<AIPlayerField>> columns = new ArrayList<ArrayList<AIPlayerField>>();
		private ArrayList<AIPlayerCombination> combinations = new ArrayList<AIPlayerCombination>();
		private AIPlayerGameState gameState;
		private final int TWO_IN_A_ROW_VALUE = 10;
		private final int THREE_IN_A_ROW_VALUE = 200;
		private final int FOUR_IN_A_ROW_VALUE = Integer.MAX_VALUE;
		private ArrayList<AIPlayerField> movesMade = new ArrayList<AIPlayerField>();

		public AIPlayerBoard(){
			initBoard();
		}

		/**
		 * Init the board
		 */
		private void initBoard(){
			for (int x = 0;x<= 6;x++) {
				ArrayList<AIPlayerField> temp = new ArrayList<AIPlayerField>();			
				for (int y = 0;y<=5;y++){					
					temp.add(new AIPlayerField());
				}				
				columns.add(temp);
			}
			// Init the possible combinations
			initCombinations();	
			// Init gamestate
			gameState = new AIPlayerGameState();
		}

		/**
		 * Check whether a column is full
		 * (Precondition 0 <= columnNumber <= 6)
		 * @param columnNumber The clumnnumber
		 * @return True if at least 1 field in the column is empty 
		 */
		public boolean isLegalMove(int columnNumber){

			if (gameState.isGameOver())
				return false;

			ArrayList<AIPlayerField> temp = columns.get(columnNumber);
			if(temp.get(5).getColor() == null)
				return true; 
			else
				return false;	
		}

		/**
		 * Get all legal moves
		 * @return An ArrayList containing Integers representing columns that are legal moves 
		 */
		public ArrayList<Integer> getLegalMoves(){
			ArrayList<Integer> result = new ArrayList<Integer>();
			for (int i = 0;i <7;i++){
				if (isLegalMove(i))
					result.add(i);
			}
			return result;		
		}

		/**
		 * Inserts a chip into the column columnNumber
		 * (Precondition 0 <= columnNumber <= 6)
		 * @param columnNumber
		 * @param color
		 * @return
		 */
		public AIPlayerField insertIntoColumn(int columnNumber, Boolean color){
			AIPlayerField result = null;
			ArrayList<AIPlayerField> temp = columns.get(columnNumber);

			for (int i = 0;i<=5;i++){
				if (temp.get(i).getColor() == null){
					temp.get(i).setColor(color);
					result = temp.get(i);
					break;
				}	
			}
			return result;
		}

		/**
		 * Make the move and update the gamestate 
		 * (Precondition 1 0 <= columnNumber <= 6)
		 * (Precondition 2 color != null)
		 * (Precondition 3 isLegalMove() == true)
		 * @param columnNumber
		 * @param color
		 */
		public void makeMove(int columnNumber, Boolean color){
			// Insert into column and get the field which received the chip
			AIPlayerField lastInserted = insertIntoColumn(columnNumber, color);
			// log the move for undo purposes
			movesMade.add(lastInserted);
			// get combos with this field
			for (AIPlayerCombination combo : combinations){
				if (combo.containsField(lastInserted))
					// Is this a winning combo?
					if (combo.isWinningCombo(color)){
						gameState.setPlayerWon(color);					
						gameState.setGameOver(true);
					}
			}
			gameState.decreaseMovesRemaining();
		}

		/**
		 * Check if the game is over. A game can be over in to ways:
		 * A player has won the game or all fields on the board are occupied by chips.
		 * @return True if game is over
		 */
		public boolean isGameOver(){
			return gameState.isGameOver();
		}

		/**
		 * Get the number of turns remaining
		 * @return Number of turns remaining
		 */
		public int getMovesRemaining(){
			return gameState.getMovesRemaining();
		}

		/**
		 * Get the current board strength for the player with the given color
		 * @param colorToMatch The color of the player. True is red, false is yellow
		 * @return An absolute score
		 */
		public int getIndividualPlayerStrength(Boolean colorToMatch){

			int score = 0;
			for (AIPlayerCombination combo : combinations){
				if (combo.isNthInARow(colorToMatch, 4)) {
					// If 4 in a row player has won
					return FOUR_IN_A_ROW_VALUE;
				} else if (combo.isNthInARow(colorToMatch, 3)) { 
					score = score +THREE_IN_A_ROW_VALUE; // If 3 in a row add a lot
				} else if (combo.isNthInARow(colorToMatch, 2)){ 
					score = score +TWO_IN_A_ROW_VALUE; // If 2 in a row add a little
				}
			}
			return score;
		}

		public int getPlayerStrength(Boolean color){
			if (color){
				if (getIndividualPlayerStrength(false) == Integer.MAX_VALUE)
					return Integer.MIN_VALUE;
				else 
					return getIndividualPlayerStrength(true) - getIndividualPlayerStrength(false);
			} else {
				if (getIndividualPlayerStrength(true) == Integer.MAX_VALUE)
					return Integer.MIN_VALUE;
				else 
					return getIndividualPlayerStrength(false) - getIndividualPlayerStrength(true);
			}
		}

		/**
		 * Undo last move on this AIPlayerBoard. This is done by fetching the last element
		 * in the movesMade arraylist, resetting AI fields color to null and removing the AI field 
		 * from the arraylist
		 */
		public void undoLastMove(){
			if (!movesMade.isEmpty()){
				AIPlayerField f = movesMade.get(movesMade.size()-1);
				f.setColor(null);
				movesMade.remove(movesMade.size()-1);
				gameState.increaseMovesRemaining();
				gameState.setGameOver(false);
				gameState.setPlayerWon(null);
			}
		}

		/**
		 * Check if a column is empty
		 * @param column The index of the column 
		 * @return True if column is empty
		 */
		public boolean isColumnEmpty(int column){			
			for (AIPlayerField f : columns.get(column)){
				if (f.getColor() != null)
					return false;
			}
			return true;
		}

		/**
		 * Initiates all the possible winning combinations and inserts these into an ArrayList
		 */
		private void initCombinations(){
			// vertical combos 21 in all
			for (ArrayList<AIPlayerField> temp : columns){
				combinations.add(new AIPlayerCombination(temp.get(0),temp.get(1),temp.get(2),temp.get(3)));
				combinations.add(new AIPlayerCombination(temp.get(1),temp.get(2),temp.get(3),temp.get(4)));
				combinations.add(new AIPlayerCombination(temp.get(2),temp.get(3),temp.get(4),temp.get(5)));
			}

			// horizontal combos 24 in all
			ArrayList<AIPlayerField> column1 = columns.get(0);
			ArrayList<AIPlayerField> column2 = columns.get(1);
			ArrayList<AIPlayerField> column3 = columns.get(2);
			ArrayList<AIPlayerField> column4 = columns.get(3);
			ArrayList<AIPlayerField> column5 = columns.get(4);
			ArrayList<AIPlayerField> column6 = columns.get(5);
			ArrayList<AIPlayerField> column7 = columns.get(6);

			for (int i = 0;i<6;i++){
				combinations.add(new AIPlayerCombination(column1.get(i),column2.get(i),column3.get(i),column4.get(i)));
				combinations.add(new AIPlayerCombination(column2.get(i),column3.get(i),column4.get(i),column5.get(i)));
				combinations.add(new AIPlayerCombination(column3.get(i),column4.get(i),column5.get(i),column6.get(i)));
				combinations.add(new AIPlayerCombination(column4.get(i),column5.get(i),column6.get(i),column7.get(i)));
			}

			// diagonal combinations 24 in all
			for (int i = 0;i<4;i++){
				ArrayList<AIPlayerField> temp1 =columns.get(i);
				ArrayList<AIPlayerField> temp2 =columns.get(i+1);
				ArrayList<AIPlayerField> temp3 =columns.get(i+2);
				ArrayList<AIPlayerField> temp4 =columns.get(i+3);
				combinations.add(new AIPlayerCombination(temp1.get(0),temp2.get(1),temp3.get(2),temp4.get(3)));
				combinations.add(new AIPlayerCombination(temp1.get(1),temp2.get(2),temp3.get(3),temp4.get(4)));
				combinations.add(new AIPlayerCombination(temp1.get(2),temp2.get(3),temp3.get(4),temp4.get(5)));		
				combinations.add(new AIPlayerCombination(temp1.get(3),temp2.get(2),temp3.get(1),temp4.get(0)));
				combinations.add(new AIPlayerCombination(temp1.get(4),temp2.get(3),temp3.get(2),temp4.get(1)));
				combinations.add(new AIPlayerCombination(temp1.get(5),temp2.get(4),temp3.get(3),temp4.get(2)));
			}
		}
	}	

/////////// Calculates the actual moves ///////////

	public class MiniMaxCalculator{

		Boolean player = false;
		Boolean adversary = true;
		int winStrength = Integer.MAX_VALUE;
		int loseStrength = Integer.MIN_VALUE;

		/**
		 * Get a new MiniMaxCalculator and set player colors
		 */
		public MiniMaxCalculator(){
			if (color){
				player = true;
				adversary = false;
			}
		}

		/**
		 * Get the next move by looking n moves ahead
		 * @param numberOfMovesToPredict The number of moves to look ahead
		 * @return The best move
		 */
		public int getNextMove(int numberOfMovesToPredict){
			// calculate max score move

			int strength = loseStrength;
			int temp = loseStrength;
			int moveToMake = -1;
			
			for (int move : b.getLegalMoves()){				
				b.makeMove(move, player);
				temp = getMin(numberOfMovesToPredict, strength);

				if (strength < temp) {
					strength = temp;
					moveToMake = move;
				}			
				b.undoLastMove();			
			}
			
			if (moveToMake == -1) {
				return b.getLegalMoves().get(0);
			}				
			return moveToMake;
		}

		/**
		 * Get the max score for this player moves
		 * @param depth The number of moves to look ahead
		 * @param parentMinimum The minimum score for the parent node (getMin in this case)
		 * @return The max score from all possible moves
		 */
		private int getMax(int depth, int parentMinimum) {

			if(depth == 0 || b.isGameOver())
				return b.getPlayerStrength(player);

			int maxStrength = loseStrength;
			for (int move : b.getLegalMoves()){				
				b.makeMove(move, player);
				int strength = getMin(depth -1, maxStrength);

				if(strength > parentMinimum) {
					b.undoLastMove();
					return strength;
				}
				if(strength > maxStrength) {
					maxStrength = strength;
				}
				b.undoLastMove();
			}
			return maxStrength;
		}

		/**
		 * Get the max score for the opponents moves
		 * @param depth The number of moves to look ahead
		 * @param parentMaximum The maximum score for the parent node (getMax in this case)
		 * @return The min score from all possible moves
		 */
		private int getMin(int depth, int parentMaximum)
		{
			if(depth == 0 || b.isGameOver())
				return b.getPlayerStrength(player);

			int minStrength = winStrength;
			for (int move : b.getLegalMoves()){				
				b.makeMove(move, adversary);
				int strength = getMax(depth -1, minStrength);
				
				if(strength < parentMaximum) {
					b.undoLastMove();
					return strength;
				}
				if(strength < minStrength) {
					minStrength = strength;
				}
				b.undoLastMove();
			}
			return minStrength;
		}
	}	
}
