package MagnusKoch;
 
public class SmartBoard
{
	private boolean gameHasBeenWon;
	public boolean redsTurn;
	private int[][] board;
	private final static int BOARD_WIDTH = 7;
	private final static int BOARD_HEIGHT = 6;
	
	private int rowNumber;
	
	private int lastMove;
	
	public int lastTokenValue; // Stores the biggest row the last token inserted is part of. Calculated when informAndCheckIfWinner is called.
	
	/**
	 * Constructs a new empty board.
	 */
	public SmartBoard()
	{
		gameHasBeenWon = false;
		redsTurn = true;
		board = new int[BOARD_WIDTH][BOARD_HEIGHT];
		lastTokenValue = 1;
			
		// The board is represented as a 2-d array of ints. 0 represents an empty position, 1 occupied by red and -1 occupied
		// by yellow.
		for(int i = 0; i < BOARD_WIDTH; i++)	
			for(int j = 0; j < BOARD_HEIGHT; j++)
				board[i][j] = 0;
	}

	/**
	 * Initializes a board with some tokens already played
	 * @param boardSituation
	 */
	public SmartBoard(int[][] boardSituation, int nextMove, boolean redsTurn)
	{
		gameHasBeenWon = false;
		this.redsTurn = redsTurn;
		this.board = boardSituation;
		informAndCheckIfWinner(nextMove);
	}
	
	public int[][] getBoardSituation()
	{
		int[][] situation = new int[BOARD_WIDTH][BOARD_HEIGHT];
		for(int i = 0; i < BOARD_WIDTH; i++)	
			for(int j = 0; j < BOARD_HEIGHT; j++)
				situation[i][j] = board[i][j];
		return situation;
	}
	
	/**
	 * The method is used for informing the board about which move has just been made. A precondition is that this move is 
	 * actually legal. The method returns true, if the move wins the game.
	 * @param columnNumber
	 * @return
	 */
	public boolean informAndCheckIfWinner(int columnNumber)
	{
		lastMove = columnNumber;
		//int rowNumber;
		int color; 
		
		if (redsTurn)
			{
				color = 1; // meaning "red"
				redsTurn = false;
			}
		else
			{
				color = -1; // meaning "yellow"
				redsTurn = true;
			}
		
	
		// Iterates through the column from bottom to top and inserts a token when encountering an empty square.
			
		for(int j = BOARD_HEIGHT -1; j >= 0; j--)
		{
					
			if(board[columnNumber][j] == 0)
			{
				board[columnNumber][j] = color; // Inserts a "color" token.
				rowNumber = j;
				gameHasBeenWon = checkForWinner(columnNumber, rowNumber, color, 1);
				break;
			}
		}
		return gameHasBeenWon;
	}
		
	
	/**
	 * Informs the board that a move has been made in the column, and remembers that move as the last.
	 * @param columnNumber
	 */
	public void inform(int columnNumber)
	{
		lastMove = columnNumber;
		int color; 
		
		if (redsTurn)
			{
				color = 1; // meaning "red"
				redsTurn = false;
			}
		else
			{
				color = -1; // meaning "yellow"
				redsTurn = true;
			}
		
		// Iterates through the column from bottom to top and inserts a token when encountering an empty square.
			
		for(int j = BOARD_HEIGHT -1; j >= 0; j--)
		{
					
			if(board[columnNumber][j] == 0)
			{
				board[columnNumber][j] = color; // Inserts a "color" token.
				break;
			}
		}
		
	}
		
	/**
	 * This method check if the token at column (i,j) is part of a winning of "5 - counter" in a row. The method return
	 * true if this is the case
	 * @param i x-coordinate of the tokens position
	 * @param j y-coordinate of the tokens position
	 * @param color the color of the token
	 * @param counter if 1, search will be for 4 in a row. If 2, search for 3 in a row.
	 * @return
	 */
	public boolean checkForWinner(int i, int j, int color, int counter)
	{
		boolean result = false;
			
		// The 1. token does not count itself.
		boolean bothDirectionsChecked = false; // The recursion goes "both ways" along a direction, initially it has not 
											  // checked both direction
			
		if (recursiveDirectionCheck(i,j, 1, 0, color, counter, bothDirectionsChecked)) result = true;
		else if	(recursiveDirectionCheck(i, j, 0, 1, color, counter, bothDirectionsChecked)) result = true;
		else if	(recursiveDirectionCheck(i,j, 1, -1, color, counter, bothDirectionsChecked)) result = true;
		else if	(recursiveDirectionCheck(i, j, 1, 1, color, counter, bothDirectionsChecked)) result = true;
		return result;
	}
		
	/**
	 * This method recursively checks if a token is in a winning combination in direction dX/dY
	 * @param i x-coordinate of the token
	 * @param j y-coordinate of the token
	 * @param dX When checking the next token the position changes dX on the x-axis
	 * @param dY When checking the next token the position changes dY on the y-axis
	 * @param color the color of the token
	 * @param counter 
	 * @param bothDirectionsChecked
	 * @return
	 */
	public boolean recursiveDirectionCheck(int i, int j, int dX, int dY, int color, int counter, boolean bothDirectionsChecked)
	{
		boolean result = false;
			
		// (If the move is not legal or the token does not have the right color) and both directions have been checked
		// we return false;
		if((!positionIsWithinBoard(i +dX, j + dY) || !(board[i + dX][j + dY] == color)) && bothDirectionsChecked) 
		{
			if(counter > lastTokenValue)
				lastTokenValue = counter;
			result = false; 
		}
		// (If the move is not legal or the token does not have the right color) and both directions have NOT been checked
		// the counter is reset, and a new recursion is started in the OPPOSITE direction.
		else if ((!positionIsWithinBoard(i +dX, j + dY) || !(board[i + dX][j + dY] == color)) && !bothDirectionsChecked) 
		{
			counter = 1;
			bothDirectionsChecked = true;
			dX = -dX;
			dY = -dY;
			result = recursiveDirectionCheck(i, j, dX, dY, color, counter, bothDirectionsChecked);
		}
		// If the counter reaches four we have a winning combination and return true. If not we check the next token in
		// the dX/dY direction.
		else
		{
			counter++;
			if(counter == 4) 
				{
					lastTokenValue = 4;
					result = true;
				}
			else
				result = recursiveDirectionCheck(i + dX, j + dY, dX, dY, color, counter, bothDirectionsChecked);
		}
		return result;
	}
		
	/**
	 * Checks if the position at (i, j) is within the boards boundaries
	 * @param i
	 * @param j
	 * @return
	 */
	public boolean positionIsWithinBoard(int i, int j)
	{
		if (i < 0 || i >= BOARD_WIDTH || j < 0 || j >= BOARD_HEIGHT) return false;
		else return true;
	}
	
	/**
	 * Evaluates the winner chance of the current board for red if color is true else for yellow.
	 * @param color
	 * @return
	 */
	public int evaluate()
	{
		if (lastTokenValue == 2) 
			return 25;
		if (lastTokenValue == 3) 
			return 50;
		else 
			return 0;
	}		
	
	public boolean gameIsOver()
	{
		return gameHasBeenWon;
	}
	
	public int getLastMove()
	{
		return lastMove;
	}
	
	public boolean spaceInColumn(int i)
	{
		if(board[i][0] == 0) 
			return true;
		else 
			return false;
	}
	
}



